use Test::More tests => 1;
BEGIN { use_ok( 'POE::Component::Server::SimpleSMTP' ) };
diag( "Testing POE::Component::Server::SimpleSMTP $POE::Component::Server::SimpleSMTP::VERSION, POE $POE::VERSION, Perl $], $^X" );
